// CFilterGraphManager.cpp: implementation of the CFilterGraphManager class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CFilterGraphManager.h"
#include "CCrossbar.h"

#include <time.h>
#include <mmreg.h>
#include <msacm.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFilterGraphManager::CFilterGraphManager()
{
	filterList = NULL;

	capGraphBuilder = NULL;
	filterGraph = NULL;
	crossbar = NULL;

	videoCapFilter = NULL;
	audioCapFilter = NULL;
	videoCompressor = NULL;
	audioCompressor = NULL;

	graphRegister = 0;
	audioInputIndex = -1;
	audioInputLevel = 0.5;
	usePreFilters = FALSE;
}

CFilterGraphManager::~CFilterGraphManager()
{

}

//////////////////////////////////////////////////////////////////////
// Helper methods
//////////////////////////////////////////////////////////////////////

int CFilterGraphManager::setFilterList(CFilterList *list)
{
	if(list == NULL)
		return -1;

	filterList = list;

	return 0;
}

int CFilterGraphManager::setVideoCaptureFilter(IBaseFilter *capFilter)
{
	if(capFilter == NULL)
		return -1;

	videoCapFilter = capFilter;

	return 0;
}

int CFilterGraphManager::setAudioCaptureFilter(IBaseFilter *capFilter)
{
	if(capFilter == NULL)
		return -1;

	audioCapFilter = capFilter;

	return 0;
}

int CFilterGraphManager::setVideoCompressor(IBaseFilter *compressor)
{
	if(compressor == NULL)
		return -1;

	videoCompressor = compressor;

	return 0;
}

int CFilterGraphManager::setAudioCompressor(IBaseFilter *compressor)
{
	if(compressor == NULL)
		return -1;

	audioCompressor = compressor;

	return 0;
}

//////////////////////////////////////////////////////////////////////
// FilterGraph methods
//////////////////////////////////////////////////////////////////////

int CFilterGraphManager::addToROT()
{
	if(graphRegister != 0 || filterGraph == NULL)
		return -1;

    IMoniker *pMoniker;
    IRunningObjectTable *pROT;
    WCHAR wsz[128];
    HRESULT hr;

    if (FAILED(GetRunningObjectTable(0, &pROT)))
        return -2;

    wsprintfW(wsz, L"FilterGraph %08x pid %08x",
		(DWORD_PTR)filterGraph, 
		GetCurrentProcessId());

    hr = CreateItemMoniker(L"!", wsz, &pMoniker);
    if (SUCCEEDED(hr)) 
    {
		hr = pROT->Register(0, filterGraph, pMoniker, &graphRegister);
        pMoniker->Release();
    }
    pROT->Release();

    if(SUCCEEDED(hr))
		return 0;
	else
		return -3;
}

int CFilterGraphManager::setVideoAmpSettings(int video_Brit, int video_Cont, int video_Hue, int video_Sat, int video_Colour)
{
	HRESULT hr;
	IAMVideoProcAmp *pAMP = NULL; 
	
	if(capGraphBuilder == NULL || videoCapFilter == NULL)
		return -1;
	
	hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video,
		videoCapFilter,
		IID_IAMVideoProcAmp,
		(void **)&pAMP);
	
	if(!pAMP)
		return -2;

	if(video_Brit > -1)
		hr = pAMP->Set(VideoProcAmp_Brightness , video_Brit, VideoProcAmp_Flags_Manual);

	if(video_Cont > -1)
		hr = pAMP->Set(VideoProcAmp_Contrast , video_Cont, VideoProcAmp_Flags_Manual);

	if(video_Hue > -1)
		hr = pAMP->Set(VideoProcAmp_Hue , video_Hue, VideoProcAmp_Flags_Manual);

	if(video_Sat > -1)
		hr = pAMP->Set(VideoProcAmp_Saturation , video_Sat, VideoProcAmp_Flags_Manual);

	hr = pAMP->Set(VideoProcAmp_ColorEnable , video_Colour, VideoProcAmp_Flags_Manual);

	
	pAMP->Release();
	pAMP = NULL;

	return 0;
}

int CFilterGraphManager::setTunerOption(int tunerInputType, int tunerCountryCode, int tunerMode)
{

	HRESULT hr;
	IAMTVTuner *pTV = NULL;
	IAMAnalogVideoDecoder *pAD = NULL;

	if(capGraphBuilder == NULL || videoCapFilter == NULL)
		return -1;

	hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video,
		videoCapFilter,
		IID_IAMTVTuner,
		(void **)&pTV);

	if(!pTV) return -2;

	hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video,
		videoCapFilter,
		IID_IAMAnalogVideoDecoder,
		(void **)&pAD);

	if(!pAD)
	{
		pTV->Release();
		return -3;
	}

	// Set the tuner up for country and input type
	// This only needs to be done once

	if(tunerInputType == 0)
		pTV->put_InputType(0, TunerInputAntenna);
	if(tunerInputType == 1)
		pTV->put_InputType(0, TunerInputCable);

	pTV->put_CountryCode(tunerCountryCode);
	pAD->put_TVFormat(tunerMode);

	pAD->Release();
	pTV->Release();

	return 0;
}

int CFilterGraphManager::setTunerChannel(int channel)
{

	HRESULT hr;
	IAMTVTuner *pTV = NULL;

	if(capGraphBuilder == NULL || videoCapFilter == NULL)
		return -1;

	hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video,
		videoCapFilter,
		IID_IAMTVTuner,
		(void **)&pTV);

	if(!pTV) return -2;

	// Change the channel
	pTV->put_Channel(channel, AMTUNER_SUBCHAN_NO_TUNE, AMTUNER_SUBCHAN_NO_TUNE);
	pTV->Release();

	return 0;
}

int CFilterGraphManager::addSmartTeeToGraph()
{
	HRESULT hr = 0;
	IBaseFilter *smartTee = NULL;

	if(usePreFilters && filterList->getFilterPointer(0))
	{
		WCHAR filterName[256];
		TCHAR filterAddNumber[256];
		filterName[0] = 0;
		
		int filterNumber = 0;
		while(filterList->getFilterPointer(filterNumber))
		{
			wsprintf(filterAddNumber, TEXT("Filter %d"), filterNumber);

#ifndef UNICODE
			MultiByteToWideChar(CP_ACP, 0, filterAddNumber, -1, filterName, NUMELMS(filterName));
#else
			wcscpy(filterName, T2W(filterAddNumber));
#endif

			ASSERT(filterList->getFilterPointer(filterNumber));

			addFilter(filterList->getFilterPointer(filterNumber), filterName);
			if(hr != S_OK)
				return -1;

			if(filterNumber == 0)
			{
				hr = capGraphBuilder->RenderStream(NULL, NULL,
					videoCapFilter, NULL, filterList->getFilterPointer(filterNumber));
			}
			else
			{
				hr = capGraphBuilder->RenderStream(NULL, NULL,
					filterList->getFilterPointer(filterNumber-1), NULL, filterList->getFilterPointer(filterNumber));
			}
			filterNumber++;
		}
		
		//Now add the Smart Tee
		CoCreateInstance(CLSID_SmartTee, NULL, CLSCTX_INPROC, IID_IBaseFilter, (void **)&smartTee);

		addFilter(smartTee, L"Smart Tea");
		if(hr != S_OK)
			return -4;

		hr = capGraphBuilder->RenderStream(NULL, NULL,
			filterList->getFilterPointer(filterNumber-1), NULL, smartTee);
	}
	else
	{
		//Now add the Smart Tee
		CoCreateInstance(CLSID_SmartTee, NULL, CLSCTX_INPROC, IID_IBaseFilter, (void **)&smartTee);

		addFilter(smartTee, L"Smart Tea");

		if(hr != S_OK)
			return -6;

		hr = capGraphBuilder->RenderStream(&PIN_CATEGORY_CAPTURE, NULL,
			videoCapFilter, NULL, smartTee);

		if(hr != S_OK)
			return -7;

	}

	return 0;
}

IBaseFilter* CFilterGraphManager::addFileWriterSection(WCHAR *outputName, BOOL compatIndex, int masterStream)
{

	IBaseFilter *fileWriter = NULL;
	IBaseFilter *mux = NULL;
	IFileSinkFilter2 *pFileWriterConfig = NULL;
	IConfigAviMux *pConfigAviMux = NULL;

	HRESULT hr = 0;
	int result = 0;

	// Create the File Writer filter and add it
    CoCreateInstance(CLSID_FileWriter,
		NULL,
		CLSCTX_INPROC,
		IID_IBaseFilter,
		(void **)&fileWriter);

	if(fileWriter == NULL)
		return NULL;

	result = addFilter(fileWriter, L"File Writer");
    if(result != 0)
		return NULL;

	// Create the AVI mux filter and add it
    CoCreateInstance(CLSID_AviDest,
		NULL, CLSCTX_INPROC,
		IID_IBaseFilter,
		(void **)&mux);

	result = addFilter(mux, L"AVI Mux");
    if(result != 0)
		return NULL;

	// Render the AVI Mux to the File Writer
	hr = capGraphBuilder->RenderStream(NULL, NULL,
		mux, NULL, fileWriter);

	if(hr != NOERROR)
		return NULL;

	// Get the FileWriter sink interface to set the overwite option
	hr = fileWriter->QueryInterface(
		IID_IFileSinkFilter2, 
		(void **)&pFileWriterConfig);

	if(!pFileWriterConfig)
	{
		return NULL;
	}

	hr = pFileWriterConfig->SetFileName(outputName, NULL);


	if(hr != NOERROR)
		return NULL;

	// Set the mode to AM_FILE_OVERWRITE
	hr = pFileWriterConfig->SetMode(AM_FILE_OVERWRITE);
	pFileWriterConfig->Release();
	pFileWriterConfig = NULL;

    // Now tell the AVIMUX to write AVI files that old apps can read properly.
    // If we don't, most apps won't be able to tell where the keyframes are,
    // slowing down editing considerably
    // Doing this will cause one seek (over the area the index will go) when
    // you capture past 1 Gig, but that's no big deal.
    // NOTE: This is on by default, so it's not necessary to turn it on

    // Also, set the proper MASTER STREAM
    hr = mux->QueryInterface(IID_IConfigAviMux,
        (void **)&pConfigAviMux);

	if(!pConfigAviMux)
	{
		return NULL;
	}

    if(hr == NOERROR && pConfigAviMux)
	{
        pConfigAviMux->SetOutputCompatibilityIndex(compatIndex);
        if(masterStream > -1)
		{
            hr = pConfigAviMux->SetMasterStream(masterStream);
        }
    }

	return mux;
}

int CFilterGraphManager::buildCaptureGraph(BOOL preview)
{
	int result = 0;
	HRESULT hr = 0;

	///////////////////////////////////////////////////////////////
	// Add and setup the Video Capture Device
	///////////////////////////////////////////////////////////////	
	if(videoCapFilter == NULL)
	{
		return -1;
	}
	else
	{
		result = addFilter(videoCapFilter, L"Video Capture Filter");
		if(result != 0)
			return -2;
	}

	// the video frame settings
	RECT *frameSize = new RECT();
	frameSize->right = 480;
	frameSize->bottom = 576;

	result = setCaptureOptions(frameSize, MEDIASUBTYPE_YUY2, 400000);
	result = addCrossBarHelper();
	result = setCrossbarIndex(1);
	result = setVideoAmpSettings(125, 125, 125, 125, 1);

	result = setTunerOption(0, 61, 16);
	result = setTunerChannel(9);

	///////////////////////////////////////////////////////////////
	// Add and setup the Audio Capture Device
	///////////////////////////////////////////////////////////////	

	if(audioCapFilter)
	{
		result = addFilter(audioCapFilter, L"Audio Capture Filter");
		if(result != 0)
			return -3;

		audioInputIndex = 5;
		result = setAudioInput();
		result = setAudioFormat(2, 44100, 176400, 4, 16, 1);

	}

	///////////////////////////////////////////////////////////////
	// Now add all the filter required to do the capture
	///////////////////////////////////////////////////////////////

	usePreFilters = FALSE;
	addSmartTeeToGraph();

	IBaseFilter *mux = addFileWriterSection(L"f:\\test.avi", FALSE, -1);

/*
	hr = capGraphBuilder->RenderStream(
		NULL,
		NULL,
		videoCapFilter,
		NULL,
		NULL);
*/
	addToROT();

//TEST
/*
    // prepare to run the graph
    IMediaControl *pMC = NULL;
    hr = filterGraph->QueryInterface(IID_IMediaControl, (void **)&pMC);

	// Run the Graph
	hr = pMC->Run();

    if(FAILED(hr))
	{
        // stop parts that started
        pMC->Stop();
        pMC->Release();
    }
  
    pMC->Release();
*/
//TEST



	return 0;
}

int CFilterGraphManager::setCrossbarIndex(int inputNumber)
{

	HRESULT hr;

	if(crossbar != NULL)
		hr = crossbar->SetInputIndex(inputNumber);

	if(hr != S_OK)
		return -1;

	return 0;
}

int CFilterGraphManager::setAudioInput()
{
	if(audioCapFilter == NULL)
		return -1;

	if(audioInputIndex > -1)
	{
		CAudioMixer *mix = new CAudioMixer(audioCapFilter);
		
		mix->setInput(audioInputIndex);
		mix->setLevel(audioInputLevel, audioInputIndex);
		
		delete mix;
	}

	return 0;
}

BOOL CFilterGraphManager::setAudioFormat(int channles, int samplesPerSec, int avgBytesPerSec, int blockAlign, int bitsPerSample, int formatTag) 
{
	if(audioCapFilter == NULL)
		return -1;

	IAMStreamConfig *audioStreamControl = NULL;

    HRESULT hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Audio,
        audioCapFilter,
		IID_IAMStreamConfig,
		(void **)&audioStreamControl);

	if(audioStreamControl == NULL)
		return -2;
	
    DWORD dwSize;
    LPWAVEFORMATEX lpwfx;
    AM_MEDIA_TYPE *pmt = NULL;
	
    // What's the largest format size around?
    acmMetrics(NULL, ACM_METRIC_MAX_SIZE_FORMAT, &dwSize);

    hr = audioStreamControl->GetFormat(&pmt);
    if(pmt == NULL || hr != NOERROR)
	{
		if(audioStreamControl)
			audioStreamControl->Release();
        return -3;
	}
	
    lpwfx = (LPWAVEFORMATEX)pmt->pbFormat;
    dwSize = max(dwSize, lpwfx->cbSize + sizeof(WAVEFORMATEX));
    lpwfx = (LPWAVEFORMATEX)GlobalAllocPtr(GHND, dwSize);

    if(lpwfx)
	{
        CopyMemory(lpwfx, pmt->pbFormat, pmt->cbFormat);
		
		lpwfx->nChannels = channles;
		lpwfx->nSamplesPerSec = samplesPerSec;
		lpwfx->nAvgBytesPerSec = avgBytesPerSec;
		lpwfx->nBlockAlign = blockAlign;
		lpwfx->wBitsPerSample = bitsPerSample;
		lpwfx->wFormatTag = formatTag;

		// while streaming
		((CMediaType *)pmt)->SetFormat((LPBYTE)lpwfx,
			lpwfx->cbSize + sizeof(WAVEFORMATEX));


		hr = audioStreamControl->SetFormat(pmt);  // filter will reconnect

		if(hr != NOERROR)
		{
			if(audioStreamControl)
				audioStreamControl->Release();
			return -4;
		}
		
        GlobalFreePtr(lpwfx) ;
    }

    DeleteMediaType(pmt);
	
	if(audioStreamControl)
		audioStreamControl->Release();
	audioStreamControl = NULL;

	return 0;
}

int CFilterGraphManager::setCaptureOptions(RECT *size, GUID colour, int framerate)
{
	if(videoCapFilter == NULL)
		return -1;

    IAMStreamConfig *videoStreamControl= NULL;

	HRESULT hr = capGraphBuilder->FindInterface(
		&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video,
		videoCapFilter,
		IID_IAMStreamConfig,
		(void **)&videoStreamControl);

	if(hr != NOERROR)
		return -2;

    AM_MEDIA_TYPE *pmt;

	int result = 0;
	
    if(videoStreamControl && videoStreamControl->GetFormat(&pmt) == S_OK)
	{
        if(pmt->formattype == FORMAT_VideoInfo)
		{
			VIDEOINFOHEADER *pvi = (VIDEOINFOHEADER *)pmt->pbFormat;
			pvi->AvgTimePerFrame = (LONGLONG)(((double)10000000) / framerate);
			pvi->bmiHeader.biWidth = size->right;
			pvi->bmiHeader.biHeight = size->bottom;
			pmt->subtype = colour;
			hr = videoStreamControl->SetFormat(pmt);

			if(hr != S_OK) result = -3;
			
        }
        DeleteMediaType(pmt);
    }	

	if(videoStreamControl)
		videoStreamControl->Release();
	videoStreamControl = NULL;

	return 0;
}

int CFilterGraphManager::addCrossBarHelper()
{
	if(videoCapFilter == NULL)
		return -1;

	IPin *pP = NULL;
	IEnumPins *pins;
	ULONG n;
	PIN_INFO pinInfo;
	BOOL Found = FALSE;
	IKsPropertySet *pKs;
	GUID guid;
	DWORD dw;
	BOOL fMatch = FALSE;
	
	crossbar = NULL;
	
	if(SUCCEEDED(videoCapFilter->EnumPins(&pins)))
	{            
		while(!Found && (S_OK == pins->Next(1, &pP, &n)))
		{
			if(S_OK == pP->QueryPinInfo(&pinInfo))
			{
				if(pinInfo.dir == PINDIR_INPUT)
				{
					// is this pin an ANALOGVIDEOIN input pin?
					if(pP->QueryInterface(IID_IKsPropertySet,(void **)&pKs) == S_OK)
					{
						if(pKs->Get(AMPROPSETID_Pin,
							AMPROPERTY_PIN_CATEGORY, NULL, 0,
							&guid, sizeof(GUID), &dw) == S_OK)
						{
							if(guid == PIN_CATEGORY_ANALOGVIDEOIN)
								fMatch = TRUE;
						}
						pKs->Release();
					}
					
					if(fMatch)
					{
						crossbar = new CCrossbar(pP);
						Found = TRUE;
					}
				}
				pinInfo.pFilter->Release();
			}
			pP->Release();
		}
		pins->Release();
	}
	
	return 0;
}

int CFilterGraphManager::initCaptureBuilder()
{

	int result = makeCaptureGraphBuilder();
	if(result != 0)
		return -1;
	
	result = makeFilterGraph();
	if(result != 0)
		return -2;

	HRESULT hr = capGraphBuilder->SetFiltergraph(filterGraph);
	if(hr != NOERROR)
		return -3;

	return 0;
}

int CFilterGraphManager::makeCaptureGraphBuilder()
{
    if(capGraphBuilder)
        return TRUE;

    HRESULT hr = 
		CoCreateInstance((REFCLSID) CLSID_CaptureGraphBuilder2,
        NULL, CLSCTX_INPROC, (REFIID) IID_ICaptureGraphBuilder2,
        (void **)&capGraphBuilder);

	if(hr == NOERROR)
		return 0;
	else
		return -1;
}

int CFilterGraphManager::makeFilterGraph() 
{
    if(filterGraph)
        return TRUE;

    HRESULT hr =
		CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC,
        IID_IGraphBuilder, (LPVOID *)&filterGraph);

	if(hr == NOERROR)
		return 0;
	else
		return -1;
}

int CFilterGraphManager::addFilter(IBaseFilter *filter, WCHAR *desc)
{

	HRESULT hr = filterGraph->AddFilter(filter, desc);

	if(hr == NOERROR)
		return 0;
	else
		return -1;
}