/*
	Virtual VCR
    Copyright (C) 2002  Shaun Faulds

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Acknowledgments:
	This application and associated filters are based on the examples
	from the Microsoft DirectX DirectShow SDK.
*/

#include <streams.h>
#include <windowsx.h>
#include <mmsystem.h>
#include "status.h"


// class names for status bar and static text windows
TCHAR szStatusClass[] = TEXT("StatusClass");
TCHAR szText[] = TEXT("SText");
int   gStatusStdHeight;   // based on font metrics

static HANDLE ghFont;
static HBRUSH ghbrHL, ghbrShadow;


//------------------------------------------------------------------------------
// Local Function Prototypes
//------------------------------------------------------------------------------
LRESULT CALLBACK statusWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK fnText(HWND, UINT, WPARAM, LPARAM);
static void PaintText(HWND hwnd, HDC hdc);



/*--------------------------------------------------------------+
| statusCreateTools - create the objects needed for status bar
|
+--------------------------------------------------------------*/
void statusCreateTools(void)
{
    HDC hdc;
    TEXTMETRIC tm;
    HFONT hfont;

    ghbrHL = CreateSolidBrush(GetSysColor(COLOR_BTNHIGHLIGHT));
    ghbrShadow = CreateSolidBrush(GetSysColor(COLOR_BTNSHADOW));

    /* Create the font we'll use for the status bar - use system as default */
    ghFont = CreateFont(12, 0,      // height, width
        0, 0,                       // escapement, orientation
        FW_NORMAL,                  // weight,
        FALSE, FALSE, FALSE,        // attributes
        ANSI_CHARSET,               // charset
        OUT_DEFAULT_PRECIS,         // output precision
        CLIP_DEFAULT_PRECIS,        // clip precision
        DEFAULT_QUALITY,            // quality
        VARIABLE_PITCH | FF_MODERN,
        TEXT("Helv"));

    if(ghFont == NULL)
    {
        ghFont = GetStockObject(SYSTEM_FONT);
    }

    // find the char size to calc standard status bar height
    hdc = GetDC(NULL);
    hfont = (HFONT)SelectObject(hdc, ghFont);
    GetTextMetrics(hdc, &tm);
    SelectObject(hdc, hfont);
    ReleaseDC(NULL, hdc);

    gStatusStdHeight = tm.tmHeight * 3 / 2;
}


/*--------------------------------------------------------------+
| statusDeleteTools
|
+--------------------------------------------------------------*/
void statusDeleteTools(void)
{
    DeleteObject(ghbrHL);
    DeleteObject(ghbrShadow);

    DeleteObject(ghFont);
}


/*--------------------------------------------------------------+
| statusInit - initialize for status window, register the
|          Window's class.
|
+--------------------------------------------------------------*/
BOOL statusInit(HANDLE hInst, HANDLE hPrev)
{
    WNDCLASS  cls;

    statusCreateTools();

    if(!hPrev)
    {
        cls.hCursor       = LoadCursor(NULL, IDC_ARROW);
        cls.hIcon         = NULL;
        cls.lpszMenuName  = NULL;
        cls.lpszClassName = szStatusClass;
        cls.hbrBackground = (HBRUSH) (COLOR_BTNFACE + 1);
        cls.hInstance     = (HINSTANCE)hInst;
        cls.style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        cls.lpfnWndProc   = (WNDPROC)statusWndProc;
        cls.cbClsExtra    = 0;
        cls.cbWndExtra    = 0;

        if(!RegisterClass(&cls))
            return FALSE;

        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = NULL;
        cls.lpszMenuName   = NULL;
        cls.lpszClassName  = szText;
        cls.hbrBackground  = (HBRUSH) (COLOR_BTNFACE + 1);
        cls.hInstance      = (HINSTANCE)hInst;
        cls.style          = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        cls.lpfnWndProc    = (WNDPROC)fnText;
        cls.cbClsExtra     = 0;
        cls.cbWndExtra     = 0;

        if(!RegisterClass(&cls))
            return FALSE;
    }

    return TRUE;
}


/*--------------------------------------------------------------+
| statusGetHeight
|
| Returns the recommended height for a status bar based on the
| character dimensions of the font used
+--------------------------------------------------------------*/
int statusGetHeight(void)
{
    return(gStatusStdHeight);
}


/*--------------------------------------------------------------+
| statusUpdateStatus - update the status line
|
| The argument can either be NULL or a string
+--------------------------------------------------------------*/
void statusUpdateStatus(HWND hwnd, LPCTSTR lpsz, int destID)
{
    HWND hwndtext = NULL;

    hwndtext = GetDlgItem(hwnd, destID);

	if(hwndtext == NULL) return;

    if(!lpsz || *lpsz == '\0')
    {
        SetWindowText(hwndtext, TEXT(""));
    }
    else
    {
        SetWindowText(hwndtext, lpsz);
    }
}


/*--------------------------------------------------------------+
| statusWndProc - window proc for status window
|
+--------------------------------------------------------------*/
LRESULT CALLBACK
statusWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    HWND hwndSText01 = NULL;
	HWND hwndSText02 = NULL;
	HWND hwndSText03 = NULL;

    switch(msg)
    {
        case WM_CREATE:
        {
            /* we need to create the static text control for the status bar */
            hwndSText01 = CreateWindow(szText,
                TEXT(""),
                WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                0, 0, 0, 0,
                hwnd,
                (HMENU) 1,  // child id
                GetWindowInstance(hwnd),
                NULL);

            if(!hwndSText01)
                return -1;

            hwndSText02 = CreateWindow(szText,
                TEXT(""),
                WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                0, 0, 0, 0,
                hwnd,
                (HMENU) 2,  // child id
                GetWindowInstance(hwnd),
                NULL);

            if(!hwndSText02)
                return -1;

            hwndSText03 = CreateWindow(szText,
                TEXT(""),
                WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
                0, 0, 0, 0,
                hwnd,
                (HMENU) 3,  // child id
                GetWindowInstance(hwnd),
                NULL);

            if(!hwndSText03)
                return -1;

            break;
        }

        case WM_DESTROY:
            statusDeleteTools();
            break;

        case WM_SIZE:
        {
            RECT rc;

            GetClientRect(hwnd, &rc);

			int top = 1;
			int bottom = rc.bottom - 2;

			int start = 5;
			int width = rc.right - 130;

            MoveWindow(GetDlgItem(hwnd, 1),    // get child window handle
                start,
				top,
                width,
                bottom,
                TRUE);

			start = rc.right - 120;
			width = 60;

            MoveWindow(GetDlgItem(hwnd, 2),    // get child window handle
                start,
				top,
                width,
				bottom,
                TRUE);

			start = rc.right - 55;
			width = 50;

            MoveWindow(GetDlgItem(hwnd, 3),    // get child window handle
                start,
				top,                   
                width,
				bottom,
                TRUE);

            break;
        }
		
        case WM_PAINT:
        {
            BeginPaint(hwnd, &ps);

            // only the background and the child window need painting
            EndPaint(hwnd, &ps);
            break;
        }

        case WM_SYSCOLORCHANGE:
            statusDeleteTools();
            statusCreateTools();
            break;

        case WM_ERASEBKGND:
            break;

    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}


/*--------------------------------------------------------------+
| fnText - window proc for static text window
|                               |
+--------------------------------------------------------------*/
LRESULT CALLBACK
fnText(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch(msg)
    {
        case WM_SETTEXT:
            DefWindowProc(hwnd, msg, wParam, lParam);
            InvalidateRect(hwnd,NULL,FALSE);
            UpdateWindow(hwnd);
            return 0L;

        case WM_ERASEBKGND:
            return 0L;

        case WM_PAINT:
        {
            PAINTSTRUCT ps;

            BeginPaint(hwnd, &ps);
            PaintText(hwnd, ps.hdc);
            EndPaint(hwnd, &ps);
            return 0L;
        }
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}


/*--------------------------------------------------------------+
| PaintText - paint the shadowed static text field
|
+--------------------------------------------------------------*/
void PaintText(HWND hwnd, HDC hdc)
{
    RECT rc;
    TCHAR ach[128];
    int  len;
    int  dx, dy;
    RECT  rcFill;
    HFONT hfontOld;
    HBRUSH hbrSave;

    GetClientRect(hwnd, &rc);

    len = GetWindowText(hwnd,ach,sizeof(ach)/sizeof(ach[0]));

    SetBkColor(hdc, GetSysColor(COLOR_BTNFACE));
    SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));

    hfontOld = (HFONT)SelectObject(hdc, ghFont);

    rcFill.left = rc.left + 1;
    rcFill.right = rc.right - 1;
    rcFill.top = rc.top + 1;
    rcFill.bottom = rc.bottom - 1;

    /* move in some and do background and text in one swoosh */
    ExtTextOut(hdc,4,1,ETO_OPAQUE,&rcFill,ach,len,NULL);

    dx = rc.right - rc.left;
    dy = rc.bottom - rc.top;

    hbrSave = (HBRUSH)SelectObject(hdc, ghbrShadow);
    PatBlt(hdc, rc.left, rc.top, 1, dy, PATCOPY);
    PatBlt(hdc, rc.left, rc.top, dx, 1, PATCOPY);

    SelectObject(hdc, ghbrHL);
    PatBlt(hdc, rc.right-1, rc.top+1, 1, dy-1, PATCOPY);
    PatBlt(hdc, rc.left+1, rc.bottom -1, dx-1, 1,  PATCOPY);

    if(hfontOld)
        SelectObject(hdc, hfontOld);
    if(hbrSave)
        SelectObject(hdc, hbrSave);
}

